/**********************************************************************
  ELFst@C̃[h
  $Id: elf_load.c,v 1.5 2004/09/18 11:23:44 tane Exp $
**********************************************************************/
#include "common.h"
#include "ppcsim.h"
#include "processor.h"
#include "funcs.h"
#include "memory.h"
#include "elf.h"

#define rev_us(a)  (a=((a&0xff00)>>8)|\
                      ((a&0x00ff)<<8))
#define rev_ui(a)  (a=((a&0xff000000)>>24)|\
		      ((a&0x00ff0000)>> 8)|\
                      ((a&0x0000ff00)<< 8)|\
                      ((a&0x000000ff)<<24))

int	elf_load( struct MACHINE_RESOURCE *mres , char *fname )
{
  int    i ;
  int    error ; 
  FILE   *ifp;
  unsigned char *mem ;
  int           mem_size ;
  
  Elf32_Ehdr  elf32_header  ;
  Elf32_Phdr *elf32_pheader ;
  Elf32_Shdr *elf32_sheader ;
  char   *name_table ;
  
  if( ( ifp = fopen( fname , "rb" )) == NULL ){
    SYSMSG("Can not open file(%s).\n",fname) ;
    return(FILE_OPEN_ERROR) ;
  }
  
  //
  // Read ELF Header
  //
  fread( &elf32_header, sizeof(Elf32_Ehdr), 1, ifp) ;
  
  if( elf32_header.e_ident[0]!=0x7f ||
      elf32_header.e_ident[1]!='E'  ||
      elf32_header.e_ident[2]!='L'  ||
      elf32_header.e_ident[3]!='F'  ){
    fclose( ifp ) ;
    SYSMSG("ELF load error.\n") ;
    return( ELF_READ_ERROR ) ;
  }
  
#ifdef HOST_ORDER_IS_LITTLE_ENDIAN
  rev_us(elf32_header.e_type) ;
  rev_us(elf32_header.e_machine) ;
  rev_ui(elf32_header.e_version) ;
  rev_ui(elf32_header.e_entry) ;
  rev_ui(elf32_header.e_phoff) ;
  rev_ui(elf32_header.e_shoff) ;
  rev_ui(elf32_header.e_flags) ;
  rev_us(elf32_header.e_ehsize) ;
  rev_us(elf32_header.e_phentsize) ;
  rev_us(elf32_header.e_phnum) ;
  rev_us(elf32_header.e_shentsize) ;
  rev_us(elf32_header.e_shnum) ;
  rev_us(elf32_header.e_shstrndx) ;
#endif /**HOST_ORDER_IS_LITTLE_ENDIAN**/
  
  //
  // Read ELF Program Header
  //
  fseek( ifp, elf32_header.e_phoff, SEEK_SET ) ;
  
  elf32_pheader=malloc( sizeof(Elf32_Phdr)*elf32_header.e_phnum ) ;
  if( elf32_pheader==NULL ){
    SYSMSG("Memory allocate error. (pheader)\n") ;
    fclose( ifp ) ;
    return(MEM_ALLOC_ERROR) ;
  }
  for( i=0 ; i<elf32_header.e_phnum ; i++ ){
    fread( &elf32_pheader[i], sizeof(Elf32_Phdr), 1, ifp) ;
    
#ifdef HOST_ORDER_IS_LITTLE_ENDIAN
    rev_ui(elf32_pheader[i].p_type ) ;
    rev_ui(elf32_pheader[i].p_offset ) ;
    rev_ui(elf32_pheader[i].p_vaddr ) ;
    rev_ui(elf32_pheader[i].p_paddr ) ;
    rev_ui(elf32_pheader[i].p_filesz ) ;
    rev_ui(elf32_pheader[i].p_memsz ) ;
    rev_ui(elf32_pheader[i].p_flags ) ;
    rev_ui(elf32_pheader[i].p_align ) ;
#endif /**HOST_ORDER_IS_LITTLE_ENDIAN**/
    
  }
  
  //
  // Read ELF Section Header
  //
  fseek( ifp, elf32_header.e_shoff, SEEK_SET ) ;
  
  elf32_sheader=malloc( sizeof(Elf32_Shdr)*elf32_header.e_shnum ) ;
  if( elf32_sheader==NULL ){
    SYSMSG("Memory allocate error. (sheader)\n") ;
    free( elf32_sheader ) ;
    fclose( ifp ) ;
    return(MEM_ALLOC_ERROR) ;
  }
  
  for( i=0 ; i<elf32_header.e_shnum ; i++ ){
    fread( &elf32_sheader[i], sizeof(Elf32_Shdr), 1, ifp) ;
    
#ifdef HOST_ORDER_IS_LITTLE_ENDIAN
    rev_ui(elf32_sheader[i].sh_name ) ;
    rev_ui(elf32_sheader[i].sh_type ) ;
    rev_ui(elf32_sheader[i].sh_flags ) ;
    rev_ui(elf32_sheader[i].sh_addr ) ;
    rev_ui(elf32_sheader[i].sh_offset ) ;
    rev_ui(elf32_sheader[i].sh_size ) ;
    rev_ui(elf32_sheader[i].sh_link ) ;
    rev_ui(elf32_sheader[i].sh_info ) ;
    rev_ui(elf32_sheader[i].sh_addralign ) ;
    rev_ui(elf32_sheader[i].sh_entsize ) ;
#endif /**HOST_ORDER_IS_LITTLE_ENDIAN**/
    
  }
  
  //
  // Read section header name table
  //
  name_table=malloc(elf32_sheader[elf32_header.e_shstrndx].sh_size) ;
  if( name_table==NULL ) {
    SYSMSG("Memory allocate error. (name_table)\n") ;
    free( elf32_pheader ) ;
    free( elf32_sheader ) ;
    fclose( ifp ) ;
    return(MEM_ALLOC_ERROR) ;
  }

  fseek( ifp, elf32_sheader[elf32_header.e_shstrndx].sh_offset, SEEK_SET ) ;
  if( fread( name_table, elf32_sheader[elf32_header.e_shstrndx].sh_size, 1, ifp)==0 ){
    free( name_table    ) ;
    free( elf32_pheader ) ;
    free( elf32_sheader ) ;
    fclose( ifp ) ;
    SYSMSG("ELF load error.\n") ;
    return( ELF_READ_ERROR ) ;
  }
  
  error=0 ;
  for( i=0 ; i<elf32_header.e_shnum ; i++ ){
    if( elf32_sheader[i].sh_type == SHT_NOBITS ) break ;
    if( elf32_sheader[i].sh_type == SHT_PROGBITS ){
      
      mem=mres->mem[seg(elf32_sheader[i].sh_addr)] ;
      mem_size=mres->mem_size[seg(elf32_sheader[i].sh_addr)] ;
      
      if( (lad(elf32_sheader[i].sh_addr)+elf32_sheader[i].sh_size) > mem_size ){
	error=1 ;
	break ;
      }
      fseek( ifp, elf32_sheader[i].sh_offset, SEEK_SET ) ;
      fread( &mem[lad(elf32_sheader[i].sh_addr)] ,
	     elf32_sheader[i].sh_size , 1, ifp) ;
    }
  }

  if( override_exec_start==NO ){
    for( i=0 ; i<elf32_header.e_shnum ; i++ ){
      if( strcmp(&name_table[elf32_sheader[i].sh_name],".text")==0 ){
	mres->pc=elf32_sheader[i].sh_addr ;
	break ;
      }
    }
  }
  
  free( name_table    ) ;
  free( elf32_pheader ) ;
  free( elf32_sheader ) ;
  fclose( ifp ) ;
  
  if( error==1 ){
    SYSMSG("ELF load error.\n") ;
    return( ELF_READ_ERROR ) ;
  }
  return( NORMAL_EXIT ) ;
}
